#
#  Copyright (C) 2005 Friedrich Leisch
#  $Id: kcca.R 1910 2005-11-17 08:22:18Z gruen $
#

normWeights <- function(x) x/mean(x)

mlogit <- function(x){
    x <- exp(-x)
    x/rowSums(x)
}
    
###**********************************************************


kccaFamily <- function(which=NULL, dist=NULL,
                       cent=NULL, name=which,
                       similarity=FALSE, preproc=NULL,
                       trim=0, groupFun="majorityClusters")
{
    if(is.null(which) && is.null(dist))
        stop("either ", sQuote("which")," or ", sQuote("dist"),
             " must be specified\n")

    if(is.null(name)) name <- deparse(substitute(dist))

    z <- new("kccaFamily", name=name, similarity=similarity)

    if(!is.null(preproc)) z@preproc <- preproc

    if(!is.null(which)){
        which <- match.arg(which, c("kmeans", "kmedians",
                                    "angle", "jaccard",
                                    "ejaccard"))
        if(!is.null(name)) z@name <- which

        if(which == "kmeans"){
            z@dist <- distEuclidean
            if(trim==0){
                z@cent <- function(x) colMeans(x)
                z@wcent <- function(x, weights)
                    colMeans(x*normWeights(weights))
            }
            else{
                z@cent <- function(x)
                    apply(x, 2, mean, trim=trim)
                z@wcent <- function(x, weights)
                    apply(x*normWeights(weights), 2, mean, trim=trim)
            }
            z@weighted <- TRUE
        }
        else if(which == "kmedians"){
            z@dist <- distManhattan
            z@cent <- function(x) apply(x, 2, median)
        }
        else if(which == "angle"){
            z@dist <- distAngle
            z@cent <- centAngle
            z@preproc <- function(x) x/sqrt(rowSums(x^2))
        }
        else if(which == "jaccard"){
            z@dist <- distJaccard
            z@cent <- function(x) centOptim01(x, dist=distJaccard)
        }
        else if(which == "ejaccard"){
            z@dist <- distJaccard
            z@cent <- function(x) colMeans(x)
        }
    }
    else{
        if(is.character(dist))
            z@dist <- get(dist, mode="function")
        else
            z@dist <- dist

        if(is.null(cent))
            z@cent <- function(x)
                centOptim(x, dist=dist)
        else if(is.character(cent))
            z@cent <- get(cent, mode="function")
        else
            z@cent <- cent        
    }

    if(similarity){
        z@cluster <- function(x, centers, n=1, distmat=NULL){

            if(is.null(distmat))
                distmat <- z@dist(x, centers)
            
            if(n==1){
                return(apply(distmat, 1, which.max))
            }
            else{
                ## <FIXME> kann ma t() einsparen?
                r <- t(matrix(apply(- distmat, 1,
                                    rank, ties.method="random"),
                              nrow=ncol(distmat)))
                ## </FIXME>
                z <- list()
                for(k in 1:n)
                    z[[k]] <- apply(r, 1, function(x) which(x==k))
            }
            return(z)
        }
    }
    else{
        z@cluster <- function(x, centers, n=1, distmat=NULL){

            if(is.null(distmat))
                distmat <- z@dist(x, centers)
            
            if(n==1){
                return(apply(distmat, 1, which.min))
            }
            else{
              r <- t(matrix(apply(distmat, 1,
                                  rank, ties.method="random"),
                            nrow=ncol(distmat)))
                z <- list()
                for(k in 1:n)
                    z[[k]] <- apply(r, 1, function(x) which(x==k))
            }
            return(z)
        }
    }

    z@allcent <- function(x, cluster, k=max(cluster, na.rm=TRUE))
    {
        centers <- matrix(NA, nrow=k, ncol=ncol(x))
        for(n in 1:k){
            if(sum(cluster==n, na.rm=TRUE)>0){
                centers[n,] <- z@cent(x[cluster==n,,drop=FALSE])
            }
        }
        centers
    }

    if(is.character(groupFun))
        groupFun <- get(groupFun, mode="function")
    z@groupFun <- groupFun
 
    
    z
}

###**********************************************************

kcca <- function(x, k, family=kccaFamily("kmeans"), weights=NULL,
                 group=NULL, control=NULL)
{
    MYCALL <- match.call()
    control <- as(control, "flexclustControl")
    x <- as(x, "matrix")
    x <- family@preproc(x)
    N <- nrow(x)
    
    if(control@classify=="auto"){
        control@classify="hard"
    }

    if(!is.null(group))
    {       
        if(length(group)>N)
            warning("group vector longer than nrow(x)")
        
        group <- rep(group, length=N)
        group <- as.factor(group)
    }
    
    if(!is.null(weights))
    {
        control@classify="weighted"

        if(!family@weighted)
            stop("Centroid function of family cannot use weights")

        if(!is.null(group))
            stop("Weighted clustering for grouped data is not implemented yet.")
        ## recycle to number of observations
        weights <- rep(weights, length=N)
    }
    
    centers <- initCenters(x, k, family)
    cluster <- centers$cluster
    k <- centers$k
    centers <- centers$centers

    sannprob <- control@simann[1]
    if(control@classify %in% c("hard", "simann"))
    {
        for(iter in 1:control@iter.max){
            
            clustold <- cluster
            distmat <- family@dist(x, centers)
            cluster <- family@cluster(x, distmat=distmat)

            if(control@classify == "simann"){
                cluster <- perturbClusters(cluster, sannprob)
                sannprob <- sannprob * control@simann[2]
            }

            if(!is.null(group))
                cluster <- family@groupFun(cluster, group, distmat)
            
            centers <- family@allcent(x, cluster=cluster, k=k)
            
            changes <- sum(cluster!=clustold)
            if(control@verbose && (iter%%control@verbose==0))
                printIter(iter, changes, "Changes")

            if(changes==0) break
        }
    }
    else if(control@classify=="weighted")
    {

        ts <- -1
        for(iter in 1:control@iter.max){

            ts.old <- ts
            distmat <- family@dist(x, centers)
            cluster <- family@cluster(distmat=distmat)

            ## for weighted kmeans we need similarities
            if(!family@similarity){
                distmat <- mlogit(distmat)
            }
            ts <- sum(distmat[cbind(1:N, cluster)])

            if(control@verbose && (iter%%control@verbose==0))
                printIter(iter, ts, "Sum of similarities")
                                
            if(abs(ts-ts.old)/(abs(ts)+0.1) < control@tolerance) break
            
            for(n in 1:k){
                w <- weights*distmat[,n]
                w[cluster==n] <- w[cluster==n]+control@gamma
                centers[n,] <- family@wcent(x, weights=w)
            }
        }
        ## make sure the final centers are centroids of their partitions
        for(n in 1:k){
            centers[n,] <- family@cent(x[cluster==n,,drop=FALSE])
        }
    }
    else
        stop("Unknown classification method")

    centers <- centers[complete.cases(centers),,drop=FALSE]
    
    newKccaObject(x=x, family=family, centers=centers, group=group,
                  iter=iter,
                  converged=(iter<control@iter.max),
                  call=MYCALL,
                  control=control)
}


initCenters <- function(x, k, family)
{
    cluster <- integer(nrow(x))
    if(is.matrix(k)){
        centers <- k
        if(any(duplicated(centers)))
            stop("initial centers are not distinct")
        k <- nrow(k)
    }
    else{
        if(length(k)>1){
            cluster <- rep(k, length=nrow(x))
            k <- max(cluster)
            centers <- family@allcent(x, cluster=cluster, k=k)
        }
        else{
            k <- as.integer(k)
            if(k<2) stop("number of clusters must be at least 2")
            ## we need to avoid duplicates here
            ux <- unique(x)
            if(nrow(ux) < k)
                stop("more centroids than distinct data points requested")
            centers <- ux[sample(1:nrow(ux), k), , drop=FALSE]
            rm(ux)
        }
    }
    list(centers=centers, cluster=cluster, k=k)
}


newKccaObject <- function(x, family, centers, group=NULL, ...)
{
    colnames(centers) <- colnames(x)

    distmat <- family@dist(x, centers)
    cluster <- family@cluster(n=2, distmat=distmat)

    if(!is.null(group))
        cluster <- family@groupFun(cluster, group, distmat)

    if(ncol(distmat)>1){
        ## at least 2 clusters
        cldist <- cbind(distmat[cbind(1:nrow(x), cluster[[1]])],
                        distmat[cbind(1:nrow(x), cluster[[2]])])
        clsim <- clusterSim(distmat,cluster)
    }
    else{
        ## only one cluster
        cldist <- distmat
        clsim <- as.matrix(1)
    }
        
    size <- as.vector(table(cluster[[1]]))
    
    clusinfo <-
        data.frame(size=size,
                   av_dist = as.vector(tapply(cldist[,1],
                                        cluster[[1]], sum))/size,
                   max_dist = as.vector(tapply(cldist[,1],
                                        cluster[[1]], max)),
                   separation = as.vector(tapply(cldist[,2],
                                        cluster[[1]], min)))

    xcent <- family@cent(x)
    totaldist <- sum(family@dist(x, matrix(xcent,nrow=1)))
    
    z <- new("kcca", k=as(nrow(centers),"integer"),
             centers=centers,
             cluster=as.integer(cluster[[1]]),
             second=as.integer(cluster[[2]]),
             family=family, 
             xrange=apply(x,2,range),
             xcent=xcent,
             totaldist=totaldist,
             clusinfo=clusinfo,
             clsim=clsim,
             cldist=cldist, ...)
    z
}
       

clusterSim <- function(distmat, cluster)
{
    K <- max(cluster[[1]])
    z <- matrix(0, ncol=K, nrow=K)

    for(k in 1:K){
        ok1 <- cluster[[1]]==k
        if(any(ok1)){
            for(n in 1:K){
                if(k!=n){
                    ok2 <- ok1 & cluster[[2]]==n
                    if(any(ok2)){
                        z[k,n] <- 2*sum(distmat[ok2,k]/
                                        (distmat[ok2,k]+distmat[ok2,n]))
                    }
                }
            }
            z[k,] <- z[k,]/sum(ok1)
        }
    }
    diag(z) <- 1
    z
}

# make random changes in cluster membership
perturbClusters <- function(cluster, prob)
{
    x <- runif(length(cluster))
    x <- (x < prob)
    if(any(x))
        cluster[x] <- sample(1:max(cluster), sum(x), replace=TRUE)

    cluster
}

## Assign each observation to the cluster where the majority from its
## group belongs to
majorityClusters <- function(cluster, group, distmat)
{
    getMax <- function(a)
    {
        b
    }

    if(is.list(cluster))
    {
        K <- max(unlist(cluster))
        ## use factors to make sure all possible clusters appear in
        ## both tables
        x <- 2*table(group, factor(cluster[[1]], levels=1:K)) +
               table(group, factor(cluster[[2]], levels=1:K))
    }
    else{
        x <- table(group, cluster)
    }

    m <- max.col(x)
    names(m) <- row.names(x)
    z <- m[group]
    names(z) <- NULL
    
    if(is.list(cluster))
    {
        ## get second best
        x[cbind(1:nrow(x), m)] <- 0
        m <- max.col(x)
        names(m) <- row.names(x)
        z1 <- m[group]
        names(z1) <- NULL
        z <- list(z, z1)
    }
    z
}
    
## Assign each observation to a cluster where no other member from its
## group belongs to
differentClusters <- function(cluster, group, distmat)
{
    if(max(table(group)) > ncol(distmat))
        stop("Number of group members must be smaller or equal as number of clusters")
       
    INF <- 2*sum(distmat, na.rm=TRUE)
    distmat[is.na(distmat)] <- INF
    if(is.list(cluster)){
        z <- getDifferentCluster(cluster[[1]], group, distmat)
        distmat[cbind(1:nrow(distmat), z)] <- INF
        cluster <- list(z,
                        getDifferentCluster(cluster[[2]], group, distmat))
    }
    else{
        cluster <- getDifferentCluster(cluster, group, distmat)
    }

    cluster    
}
    
getDifferentCluster <- function(cluster, group, distmat)
{
    x <- table(group, cluster)
    ok <- (apply(x, 1, max)==1)
    nok.names <- unique(row.names(x[!ok,,drop=FALSE]))
    
    for(n in nok.names){
        ok <- group==n
        if(sum(ok)>1)
            cluster[ok] <- solve_LSAP1(distmat[ok,])
    }
    cluster
}

solve_LSAP1 <- function (x, maximum = FALSE) 
{
    require("clue")
    if (!is.matrix(x) || any(x < 0)) 
        stop("x must be a matrix with nonnegative entries.")

    if(nrow(x)>ncol(x))
        stop("x must have less or equal rows than columns")

    nr <- nrow(x)
    nc <- ncol(x)
    if(ncol(x) > nrow(x))
        x <- rbind(x, matrix(2*sum(x), nrow=(ncol(x)-nrow(x)), ncol=ncol(x)))

    if (maximum)
        x <- max(x) - x
    
    storage.mode(x) <- "double"
    out <- .C("solve_LSAP", x, nc, p = integer(nc), PACKAGE = "clue")$p + 1
    out <- out[1:nr]
    class(out) <- "solve_LSAP"
    out
}    

###**********************************************************

setMethod("show", "kcca",
function(object)
{
    cat("kcca object of family", sQuote(object@family@name),"\n\n")
    cat("call:", deparse(object@call,0.75*getOption("width")),
        sep="\n")
    cat("\ncluster sizes:\n")
    print(table(cluster(object), exclude=NULL))
    cat("\n")
})
        
setMethod("summary", "kcca",
function(object)
{
    cat("kcca object of family", sQuote(object@family@name),"\n\n")
    cat("call:", deparse(object@call,0.75*getOption("width")),
        sep="\n")
    cat("\ncluster info:\n")
    print(object@clusinfo)
    cat("\n")
    if(!object@converged) cat("no ")
    cat("convergence after", object@iter, "iterations\n")
    cat("sum of within cluster distances:", info(object, "distsum"),"\n")
})

###**********************************************************


setGeneric("neighbours",
function(object, symm=TRUE, ...) standardGeneric("neighbours"))

setMethod("neighbours", signature(object="kcca"),
function(object, symm=TRUE, ...)
{
    neighbours(list(object@cluster, object@second))
})

setMethod("neighbours", signature(object="list"),
function(object, symm=TRUE, ...)
{
    z <- list()
    for(n in 1:max(object[[1]], object[[2]])){
        if(symm)
            z[[n]] <- unique(c(object[[2]][object[[1]]==n],
                               object[[1]][object[[2]]==n]))
        else
            z[[n]] <- unique(object[[2]][object[[1]]==n])
    }
    z
})

###**********************************************************


setGeneric("ntable", function(object, ...)
           standardGeneric("ntable"))

setMethod("ntable", signature(object="kcca"),
function(object, freq=TRUE, distance=FALSE, ...)
{
    z <- table(object@cluster, object@second)

    if(distance){
        freq <- FALSE
        z[z==0] <- NA
        z <- length(object@cluster)-z
        diag(z) <- 0
    }
    
    if(!freq)
        z <- z/rowSums(z,na.rm=TRUE)

    z
})
        
    
###**********************************************************        

stepFlexclust <- function(..., K=NULL, nrep=3, verbose=TRUE,
                          FUN=kcca, drop=TRUE)
{
    MYCALL <- match.call()
    
    bestKcca <- function(...)
    {
        distsum <- Inf
        for(m in 1:nrep){
            if(verbose) cat(" *")
            x = FUN(...)
            newdistsum <- info(x, "distsum")
            if(newdistsum < distsum){
              z <- x
              distsum <- newdistsum
            }
        }
        z
    }

    if(is.null(K)){
        z = bestKcca(...)
        z@call <- MYCALL
        if(verbose) cat("\n")
        return(z)
    }
    
    K = as.integer(K)
    if(length(K)==0)
        return(list())
    
    z = list()
    MYCALL1 <- MYCALL
    if("drop" %in% names(MYCALL))
        MYCALL1[["drop"]] <- NULL
    
    for(n in 1:length(K)){
        if(verbose) cat(K[n], ":")
        z[[as.character(K[n])]] = bestKcca(..., k=K[n])
        MYCALL1[["K"]] <- K[n]
        
        z[[as.character(K[n])]]@call <- MYCALL1
        if(verbose) cat("\n")
    }
    if(drop && length(K)==1){
        return(z[[1]])
    }
    else{
        return(new("stepFlexclust", models=z, K=as(K, "integer"),
                   nrep=as(nrep, "integer"), call=MYCALL))
    }
}


setMethod("show", "stepFlexclust",
function(object)
{
    cat("stepFlexclust object of family",
        sQuote(object@models[[1]]@family@name),"\n\n")
    cat("call:", deparse(object@call,0.75*getOption("width")),
        sep="\n")
    cat("\n")
    
    z <- data.frame(iter = sapply(object@models, function(x) x@iter),
                    converged = sapply(object@models, function(x) x@converged),
                    distsum = sapply(object@models,
                                     function(x) info(x, "distsum")))
    z1 <- data.frame(iter = NA,
                     converged = NA,
                     distsum = object@models[[1]]@totaldist)
    z <- rbind(z1, z)
    
    print(z, na.string="")
})
    
setGeneric("getModel",
function(object, ...) standardGeneric("getModel"))

setMethod("getModel", "stepFlexclust",
function(object, which=1)
{
    object@models[[which]]
})
